package rescuecore.mas08;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Scanner;

import rescuecore.PlatoonAgent;
import rescuecore.RescueConstants;
import rescuecore.RescueObject;
import rescuecore.event.MemoryListener;
import rescuecore.event.ObjectAddedEvent;
import rescuecore.event.ObjectChangedEvent;
import rescuecore.mas08.ExplorationCivilian.CIVILIAN_STATES;
import rescuecore.mas08.ExplorationCivilian.SENSE_STATE;
import rescuecore.objects.AmbulanceTeam;
import rescuecore.objects.Civilian;
import rescuecore.objects.MotionlessObject;
import rescuecore.sample.SampleSearch;

/**
 * A simplified agent used for exploration only.
 *
 * There is a list of possible search targets and the agent can provide 
 * one of these to move to.  
 *
 * The simplified world model reports Civilians, that have been found.
 * 
 * In addition to the RescueSimulation architecture civilians also
 * provide different signs of life, which can be sensed dependent on each
 * agents capabilities.
 */
public abstract class BaseExplorationAgent extends PlatoonAgent {
	/**
	 * The following data structures shall be used in an
	 * ExplorationAgent.
	 */
	protected List<ExplorationTarget> explorationTargets;
	protected List<ExplorationCivilian> explorationCivilians;
	
	/**
	 * The next data describes the ground truth, provided by data files.
	 * Do NOT access it in agents' code, this is the information
	 * agents shall determine by exploration.
	 */

	/**
	 * Map civilians by id to their ground truth state.
	 */
	private Map<Integer, CIVILIAN_STATES> groundTruthStates;
	private String currentRunningMap;

	/**
	 * Pass arguments to agent to get the sensors as: T M S C
	 */
	private boolean hasThermalSensor;
	private boolean hasMotionSensor;
	private boolean hasSoundSensor;
	private boolean hasCo2Sensor;
	
	public BaseExplorationAgent(String [] args) {
		super(RescueConstants.AGENT_TYPE_AMBULANCE_TEAM);
		currentRunningMap = args[0];
		hasThermalSensor = false;
		hasMotionSensor = false;
		hasSoundSensor = false;
		hasCo2Sensor = false;
		for(int i = 1; i < args.length; i++) {
			if(args[i].equals("T")) {
				hasThermalSensor = true;
			} else if(args[i].equals("M")) {
				hasMotionSensor = true;
			} else if(args[i].equals("S")) {
				hasSoundSensor = true;
			} else if(args[i].equals("C")) {
				hasCo2Sensor = true;
			}
		}
		System.out.println("Agent equipped with " 
				+ (hasThermalSensor ? "T " : "") 
				+ (hasMotionSensor ? "M " : "")
				+ (hasSoundSensor ? "S" : "")
				+ (hasCo2Sensor ? "C" : ""));
		if(!hasThermalSensor && !hasMotionSensor && !hasSoundSensor && !hasCo2Sensor) {
			System.err.println("No sensors equipped - I cant do anything.");
		}
	}

	
    /**
       Get a reference to the AmbulanceTeam controlled by this agent
       @return the AmbulanceTeam controlled by this agent
    */
    protected AmbulanceTeam me() {
		return (AmbulanceTeam)memory.lookup(id);
    }
    

    public void initialise(RescueObject[] knowledge, RescueObject self) {
		// Add a memory listener so that we get informed about changes to humanoids
		super.initialise(knowledge,self);

		memory.addMemoryListener(new MemoryListener() {
				
				private void advancedSense(Civilian c) {
					CIVILIAN_STATES cs = groundTruthStates.get(new Integer(c.getID()));
					if(cs == null)
						return;
					ExplorationCivilian expl = null;
					for(ExplorationCivilian next : explorationCivilians) {
						if(next.civilian == c) {
							expl = next;
							break;
						}
					}
					// this is a new one
					if(expl == null) {
						expl = new ExplorationCivilian(c);
						explorationCivilians.add(expl);
					}
					
					// TODO: Later here with some probability: Concious ones may not move or sound.					
					if(hasThermalSensor) {
						expl.thermal = SENSE_STATE.SS_TRUE;
					}
					if(hasMotionSensor) {
						if(cs == CIVILIAN_STATES.CS_CONCIOUS) {
							expl.motion = SENSE_STATE.SS_TRUE;
						} else if(cs == CIVILIAN_STATES.CS_UNCONCIOUS) {
							expl.motion = SENSE_STATE.SS_FALSE;
						}
					}
					if(hasSoundSensor) {
						if(cs == CIVILIAN_STATES.CS_CONCIOUS) {
							expl.sound = SENSE_STATE.SS_TRUE;
						} else if(cs == CIVILIAN_STATES.CS_UNCONCIOUS) {
							expl.sound = SENSE_STATE.SS_FALSE;
						}
					}
					if(hasCo2Sensor) {
						expl.co2 = SENSE_STATE.SS_TRUE;
					}
				}

				public void objectAdded(ObjectAddedEvent event) {
					RescueObject o = event.getObject();
					if (o.isCivilian()) {
						Civilian c = (Civilian)o;
						advancedSense(c);
					}
				}
				
				public void objectChanged(ObjectChangedEvent event) {
					RescueObject o = event.getObject();
					if (o.isCivilian()) {
						Civilian c = (Civilian)o;
						advancedSense(c);
					}
				}
			});

		// init data structures
		groundTruthStates = new HashMap<Integer, CIVILIAN_STATES>();
		explorationTargets = new ArrayList<ExplorationTarget>();
		explorationCivilians = new ArrayList<ExplorationCivilian>();
		
		String kernelHome = System.getenv("RESCUE_KERNEL_HOME");
		if(kernelHome == null) {
			System.err.println("No RESCUE_KERNEL_HOME environment variable defined - cannot work.");
			System.exit(1);
		} else {
			System.out.println("Kernel home is " + kernelHome);
			File kernelFile = new File(kernelHome);
			File mapsFile = new File(kernelFile, "maps");
			File curMapFile = new File(mapsFile, currentRunningMap);
			
			File civFile = new File(curMapFile, "civstates.txt");
			File targetsFile = new File(curMapFile, "targetlocs.txt");

			Scanner s;
			try {
				s = new Scanner(civFile);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
				System.exit(1);
				return;
			}
			while(s.hasNextLine()) {
				String ss = s.nextLine();
				parseCivState(ss);
			}

			Scanner targetsScanner;
			try {
				targetsScanner = new Scanner(targetsFile);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
				System.exit(1);
				return;
			}
			while(targetsScanner.hasNextLine()) {
				parseTargetLoc(targetsScanner.nextLine());
			}
			
			//System.out.println(groundTruthStates.toString());
			//System.out.println(explorationTargets);
		}
    }

	private void parseTargetLoc(String l) {
		l = l.trim();
		if(l.length()==0)
			return;
		int locId = Integer.valueOf(l).intValue();
		RescueObject obj = memory.lookup(locId);
		if(obj == null)
			return;
		if(obj instanceof MotionlessObject) {
			MotionlessObject mo = (MotionlessObject) obj;
			ExplorationTarget t = new ExplorationTarget(mo);
			explorationTargets.add(t);
		}			
	}


	private void parseCivState(String l) {
		if(l.length()==0)
			return;
		if(l.startsWith("#"))
			return;
		l = l.trim();
		String [] entries = l.split(" ");
		if(entries.length != 2)	{
			System.err.println("Found bad civstate line: " + l);
			return;
		}
		int id = Integer.valueOf(entries[0]).intValue();
		Integer idInt = new Integer(id);
		CIVILIAN_STATES state = CIVILIAN_STATES.CS_UNKNOWN;
		String stateStr = entries[1];
		//System.out.println("stateStr: " + stateStr );
		if(stateStr.equals("U")) {
			//System.out.println("unconcious");
			state = CIVILIAN_STATES.CS_UNCONCIOUS;
		} else if(stateStr.equals("C")) {
			//System.out.println("concious");
			state = CIVILIAN_STATES.CS_CONCIOUS;
		} else {
			System.err.println("Unknown civilian state: " + stateStr);
		}
		
		groundTruthStates.put(idInt, state);
	}

	protected void moveTo(ExplorationTarget t) {
		int[] path = SampleSearch.breadthFirstSearch(getLocation(), t.target, memory); // Find a path from my current location to the target's location
		if (path!=null) {
			//				System.out.println(this+" moving to " + t.target);
			move(path);
		}
	}

}
